/* --------------------------------------------------------------------------------------------
 * Copyright (c) Remy Suen. All rights reserved.
 * Licensed under the MIT License. See License.txt in the project root for license information.
 * ------------------------------------------------------------------------------------------ */
'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
const vscode_languageserver_1 = require("vscode-languageserver");
const dockerfile_utils_1 = require("dockerfile-utils");
const dockerfile_language_service_1 = require("dockerfile-language-service");
/**
 * The settings to use for the validator if the client doesn't support
 * workspace/configuration requests.
 */
let validatorSettings = null;
/**
 * The validator settings that correspond to an individual file retrieved via
 * the workspace/configuration request.
 */
let validatorConfigurations = new Map();
let connection = vscode_languageserver_1.createConnection();
let service = dockerfile_language_service_1.DockerfileLanguageServiceFactory.createLanguageService();
service.setLogger({
    log(message) {
        connection.console.log(message);
    }
});
/**
 * Whether the client supports the workspace/applyEdit request.
 */
let applyEditSupport = false;
/**
 * Whether the client supports the workspace/configuration request.
 */
let configurationSupport = false;
let documentChangesSupport = false;
let documents = {};
function supportsSnippets(capabilities) {
    return capabilities.textDocument
        && capabilities.textDocument.completion
        && capabilities.textDocument.completion.completionItem
        && capabilities.textDocument.completion.completionItem.snippetSupport;
}
function getHoverContentFormat(capabilities) {
    return capabilities.textDocument
        && capabilities.textDocument.hover
        && capabilities.textDocument.hover.contentFormat;
}
function setServiceCapabilities(capabilities) {
    service.setCapabilities({
        completion: {
            completionItem: {
                snippetSupport: supportsSnippets(capabilities)
            }
        },
        hover: {
            contentFormat: getHoverContentFormat(capabilities)
        }
    });
}
connection.onInitialized(() => {
    if (configurationSupport) {
        // listen for notification changes if the client supports workspace/configuration
        connection.client.register(vscode_languageserver_1.DidChangeConfigurationNotification.type);
    }
});
connection.onInitialize((params) => {
    setServiceCapabilities(params.capabilities);
    applyEditSupport = params.capabilities.workspace && params.capabilities.workspace.applyEdit === true;
    documentChangesSupport = params.capabilities.workspace && params.capabilities.workspace.workspaceEdit && params.capabilities.workspace.workspaceEdit.documentChanges === true;
    configurationSupport = params.capabilities.workspace && params.capabilities.workspace.configuration === true;
    return {
        capabilities: {
            textDocumentSync: vscode_languageserver_1.TextDocumentSyncKind.Incremental,
            codeActionProvider: applyEditSupport,
            completionProvider: {
                resolveProvider: true,
                triggerCharacters: [
                    '=',
                    ' ',
                    '$',
                    '-',
                ]
            },
            executeCommandProvider: applyEditSupport ? {
                commands: [
                    dockerfile_language_service_1.CommandIds.LOWERCASE,
                    dockerfile_language_service_1.CommandIds.UPPERCASE,
                    dockerfile_language_service_1.CommandIds.EXTRA_ARGUMENT,
                    dockerfile_language_service_1.CommandIds.DIRECTIVE_TO_BACKSLASH,
                    dockerfile_language_service_1.CommandIds.DIRECTIVE_TO_BACKTICK,
                    dockerfile_language_service_1.CommandIds.FLAG_TO_CHOWN,
                    dockerfile_language_service_1.CommandIds.FLAG_TO_COPY_FROM,
                    dockerfile_language_service_1.CommandIds.FLAG_TO_HEALTHCHECK_INTERVAL,
                    dockerfile_language_service_1.CommandIds.FLAG_TO_HEALTHCHECK_RETRIES,
                    dockerfile_language_service_1.CommandIds.FLAG_TO_HEALTHCHECK_START_PERIOD,
                    dockerfile_language_service_1.CommandIds.FLAG_TO_HEALTHCHECK_TIMEOUT,
                    dockerfile_language_service_1.CommandIds.CONVERT_TO_AS,
                    dockerfile_language_service_1.CommandIds.REMOVE_EMPTY_CONTINUATION_LINE
                ]
            } : undefined,
            documentFormattingProvider: true,
            documentRangeFormattingProvider: true,
            documentOnTypeFormattingProvider: {
                firstTriggerCharacter: '\\',
                moreTriggerCharacter: ['`']
            },
            hoverProvider: true,
            documentSymbolProvider: true,
            documentHighlightProvider: true,
            renameProvider: true,
            definitionProvider: true,
            signatureHelpProvider: {
                triggerCharacters: [
                    '-',
                    '[',
                    ',',
                    ' ',
                    '='
                ]
            },
            documentLinkProvider: {
                resolveProvider: false
            }
        }
    };
});
function validateTextDocument(textDocument) {
    if (configurationSupport) {
        getConfiguration(textDocument.uri).then((config) => {
            let maintainer = dockerfile_utils_1.ValidationSeverity.WARNING;
            let directiveCasing = dockerfile_utils_1.ValidationSeverity.WARNING;
            let emptyContinuationLine = dockerfile_utils_1.ValidationSeverity.WARNING;
            let instructionCasing = dockerfile_utils_1.ValidationSeverity.WARNING;
            let instructionCmdMultiple = dockerfile_utils_1.ValidationSeverity.WARNING;
            let instructionEntrypointMultiple = dockerfile_utils_1.ValidationSeverity.WARNING;
            let instructionHealthcheckMultiple = dockerfile_utils_1.ValidationSeverity.WARNING;
            if (config) {
                maintainer = getSeverity(config.deprecatedMaintainer);
                directiveCasing = getSeverity(config.directiveCasing);
                emptyContinuationLine = getSeverity(config.emptyContinuationLine);
                instructionCasing = getSeverity(config.instructionCasing);
                instructionCmdMultiple = getSeverity(config.instructionCmdMultiple);
                instructionEntrypointMultiple = getSeverity(config.instructionEntrypointMultiple);
                instructionHealthcheckMultiple = getSeverity(config.instructionHealthcheckMultiple);
            }
            const fileSettings = {
                deprecatedMaintainer: maintainer,
                directiveCasing: directiveCasing,
                emptyContinuationLine: emptyContinuationLine,
                instructionCasing: instructionCasing,
                instructionCmdMultiple: instructionCmdMultiple,
                instructionEntrypointMultiple: instructionEntrypointMultiple,
                instructionHealthcheckMultiple: instructionHealthcheckMultiple
            };
            const diagnostics = service.validate(textDocument.getText(), fileSettings);
            connection.sendDiagnostics({ uri: textDocument.uri, diagnostics });
        });
    }
    else {
        const diagnostics = service.validate(textDocument.getText(), validatorSettings);
        connection.sendDiagnostics({ uri: textDocument.uri, diagnostics });
    }
}
function getSeverity(severity) {
    switch (severity) {
        case "ignore":
            return dockerfile_utils_1.ValidationSeverity.IGNORE;
        case "warning":
            return dockerfile_utils_1.ValidationSeverity.WARNING;
        case "error":
            return dockerfile_utils_1.ValidationSeverity.ERROR;
    }
    return null;
}
/**
 * Gets the validation configuration that pertains to the specified resource.
 *
 * @param resource the interested resource
 * @return the configuration to use to validate the interested resource
 */
function getConfiguration(resource) {
    let result = validatorConfigurations.get(resource);
    if (!result) {
        result = connection.workspace.getConfiguration({ section: "docker.languageserver.diagnostics", scopeUri: resource });
        validatorConfigurations.set(resource, result);
    }
    return result;
}
// listen for notifications when the client's configuration has changed
connection.onNotification(vscode_languageserver_1.DidChangeConfigurationNotification.type, () => {
    refreshConfigurations();
});
/**
 * Wipes and reloads the internal cache of validator configurations.
 */
function refreshConfigurations() {
    // store all the URIs that need to be refreshed
    const settingsRequest = [];
    for (let uri in documents) {
        settingsRequest.push({ section: "docker.languageserver.diagnostics", scopeUri: uri });
    }
    // clear the cache
    validatorConfigurations.clear();
    // ask the workspace for the configurations
    connection.workspace.getConfiguration(settingsRequest).then((values) => {
        const toRevalidate = [];
        for (let i = 0; i < values.length; i++) {
            const resource = settingsRequest[i].scopeUri;
            // a value might have been stored already, use it instead and ignore this one if so
            if (values[i] && !validatorConfigurations.has(resource)) {
                validatorConfigurations.set(resource, Promise.resolve(values[i]));
                toRevalidate.push(resource);
            }
        }
        for (const resource of toRevalidate) {
            validateTextDocument(documents[resource]);
        }
    });
}
connection.onDidChangeConfiguration((change) => {
    if (configurationSupport) {
        refreshConfigurations();
    }
    else {
        let settings = change.settings;
        let maintainer = dockerfile_utils_1.ValidationSeverity.WARNING;
        let directiveCasing = dockerfile_utils_1.ValidationSeverity.WARNING;
        let emptyContinuationLine = dockerfile_utils_1.ValidationSeverity.WARNING;
        let instructionCasing = dockerfile_utils_1.ValidationSeverity.WARNING;
        let instructionCmdMultiple = dockerfile_utils_1.ValidationSeverity.WARNING;
        let instructionEntrypointMultiple = dockerfile_utils_1.ValidationSeverity.WARNING;
        let instructionHealthcheckMultiple = dockerfile_utils_1.ValidationSeverity.WARNING;
        if (settings.docker && settings.docker.languageserver && settings.docker.languageserver.diagnostics) {
            maintainer = getSeverity(settings.docker.languageserver.diagnostics.deprecatedMaintainer);
            directiveCasing = getSeverity(settings.docker.languageserver.diagnostics.directiveCasing);
            emptyContinuationLine = getSeverity(settings.docker.languageserver.diagnostics.emptyContinuationLine);
            instructionCasing = getSeverity(settings.docker.languageserver.diagnostics.instructionCasing);
            instructionCmdMultiple = getSeverity(settings.docker.languageserver.diagnostics.instructionCmdMultiple);
            instructionEntrypointMultiple = getSeverity(settings.docker.languageserver.diagnostics.instructionEntrypointMultiple);
            instructionHealthcheckMultiple = getSeverity(settings.docker.languageserver.diagnostics.instructionHealthcheckMultiple);
        }
        validatorSettings = {
            deprecatedMaintainer: maintainer,
            directiveCasing: directiveCasing,
            emptyContinuationLine: emptyContinuationLine,
            instructionCasing: instructionCasing,
            instructionCmdMultiple: instructionCmdMultiple,
            instructionEntrypointMultiple: instructionEntrypointMultiple,
            instructionHealthcheckMultiple: instructionHealthcheckMultiple
        };
        // validate all the documents again
        Object.keys(documents).forEach((key) => {
            validateTextDocument(documents[key]);
        });
    }
});
connection.onCompletion((textDocumentPosition) => {
    let document = documents[textDocumentPosition.textDocument.uri];
    if (document) {
        return service.computeCompletionItems(document.getText(), textDocumentPosition.position);
    }
    return null;
});
connection.onSignatureHelp((textDocumentPosition) => {
    let document = documents[textDocumentPosition.textDocument.uri];
    if (document !== null) {
        return service.computeSignatureHelp(document.getText(), textDocumentPosition.position);
    }
    return {
        signatures: [],
        activeSignature: null,
        activeParameter: null,
    };
});
connection.onCompletionResolve((item) => {
    return service.resolveCompletionItem(item);
});
connection.onHover((textDocumentPosition) => {
    let document = documents[textDocumentPosition.textDocument.uri];
    if (document !== null) {
        return service.computeHover(document.getText(), textDocumentPosition.position);
    }
    return null;
});
connection.onDocumentHighlight((textDocumentPosition) => {
    let document = documents[textDocumentPosition.textDocument.uri];
    if (document) {
        return service.computeHighlightRanges(document.getText(), textDocumentPosition.position);
    }
    return [];
});
connection.onCodeAction((codeActionParams) => {
    if (applyEditSupport && codeActionParams.context.diagnostics.length > 0) {
        return service.computeCodeActions(codeActionParams.textDocument, codeActionParams.range, codeActionParams.context);
    }
    return [];
});
connection.onExecuteCommand((params) => {
    if (applyEditSupport) {
        let uri = params.arguments[0];
        let document = documents[uri];
        if (document) {
            let edits = service.computeCommandEdits(document.getText(), params.command, params.arguments);
            if (edits) {
                if (documentChangesSupport) {
                    let identifier = vscode_languageserver_1.VersionedTextDocumentIdentifier.create(uri, document.version);
                    connection.workspace.applyEdit({
                        documentChanges: [
                            vscode_languageserver_1.TextDocumentEdit.create(identifier, edits)
                        ]
                    });
                }
                else {
                    connection.workspace.applyEdit({
                        changes: {
                            [uri]: edits
                        }
                    });
                }
            }
        }
    }
});
connection.onDefinition((textDocumentPosition) => {
    let uri = textDocumentPosition.textDocument.uri;
    let document = documents[uri];
    if (document) {
        return service.computeDefinition(textDocumentPosition.textDocument, document.getText(), textDocumentPosition.position);
    }
    return null;
});
connection.onRenameRequest((params) => {
    let document = documents[params.textDocument.uri];
    if (document) {
        let edits = service.computeRename(params.textDocument, document.getText(), params.position, params.newName);
        return {
            changes: {
                [params.textDocument.uri]: edits
            }
        };
    }
    return null;
});
connection.onDocumentSymbol((documentSymbolParams) => {
    let uri = documentSymbolParams.textDocument.uri;
    let document = documents[uri];
    if (document) {
        return service.computeSymbols(documentSymbolParams.textDocument, document.getText());
    }
    return [];
});
connection.onDocumentFormatting((documentFormattingParams) => {
    let document = documents[documentFormattingParams.textDocument.uri];
    if (document) {
        return service.format(document.getText(), documentFormattingParams.options);
    }
    return [];
});
connection.onDocumentRangeFormatting((rangeFormattingParams) => {
    let document = documents[rangeFormattingParams.textDocument.uri];
    if (document) {
        return service.formatRange(document.getText(), rangeFormattingParams.range, rangeFormattingParams.options);
    }
    return [];
});
connection.onDocumentOnTypeFormatting((onTypeFormattingParams) => {
    const document = documents[onTypeFormattingParams.textDocument.uri];
    if (document) {
        return service.formatOnType(document.getText(), onTypeFormattingParams.position, onTypeFormattingParams.ch, onTypeFormattingParams.options);
    }
    return [];
});
connection.onDocumentLinks((documentLinkParams) => {
    let document = documents[documentLinkParams.textDocument.uri];
    if (document) {
        return service.computeLinks(document.getText());
    }
    return null;
});
connection.onDidOpenTextDocument((didOpenTextDocumentParams) => {
    let document = vscode_languageserver_1.TextDocument.create(didOpenTextDocumentParams.textDocument.uri, didOpenTextDocumentParams.textDocument.languageId, didOpenTextDocumentParams.textDocument.version, didOpenTextDocumentParams.textDocument.text);
    documents[didOpenTextDocumentParams.textDocument.uri] = document;
    validateTextDocument(document);
});
function getLaterChange(changes, i, j) {
    if (changes[i].range.start.line === changes[j].range.start.line) {
        return changes[i].range.start.character < changes[j].range.start.character ? j : i;
    }
    else if (changes[i].range.start.line < changes[j].range.start.line) {
        return j;
    }
    return i;
}
function sortChanges(changes) {
    let sorted = [];
    let length = changes.length;
    for (let i = 0; i < length; i++) {
        let candidate = 0;
        for (let j = 1; j < changes.length; j++) {
            candidate = getLaterChange(changes, candidate, j);
        }
        sorted.push(changes[candidate]);
        changes.splice(candidate, 1);
    }
    return sorted;
}
function handleChanges(document, content, changes) {
    if (changes.length === 1 && !changes[0].range) {
        // not an incremental change
        return changes[0].text;
    }
    else if (changes.length !== 0) {
        changes = sortChanges(changes);
        for (let i = 0; i < changes.length; i++) {
            let offset = document.offsetAt(changes[i].range.start);
            let end = null;
            if (changes[i].range.end) {
                end = document.offsetAt(changes[i].range.end);
            }
            else {
                end = offset + changes[i].rangeLength;
            }
            content = content.substring(0, offset) + changes[i].text + content.substring(end);
        }
    }
    return content;
}
connection.onDidChangeTextDocument((didChangeTextDocumentParams) => {
    let document = documents[didChangeTextDocumentParams.textDocument.uri];
    let buffer = document.getText();
    let changes = didChangeTextDocumentParams.contentChanges;
    let changed = handleChanges(document, buffer, changes);
    if (changed !== buffer) {
        document = vscode_languageserver_1.TextDocument.create(didChangeTextDocumentParams.textDocument.uri, document.languageId, didChangeTextDocumentParams.textDocument.version, changed);
        documents[didChangeTextDocumentParams.textDocument.uri] = document;
        validateTextDocument(document);
    }
});
connection.onDidCloseTextDocument((didCloseTextDocumentParams) => {
    validatorConfigurations.delete(didCloseTextDocumentParams.textDocument.uri);
    delete documents[didCloseTextDocumentParams.textDocument.uri];
});
// setup complete, start listening for a client connection
connection.listen();
